#!/usr/bin/perl -w
#
# treePruner.pl v1.0
# Last modified June 16, 2005 11:46
# (c) Olaf R.P. Bininda-Emonds
#
# Input:
#   1) A nexus-formatted treefile containing one or more tree descriptions.
#	2) (Optional) A list of names (one per line) to be pruned from the input tree descriptions. Can automatically read in output from STRindexer.pl
#
# Output:
#   A nexus-formatted treefile with the input tree descriptions pruned of specifically named taxa and/or to the taxa that they have in common. Can optionally maintain support values for non-pruned nodes.
#
# Requires:
#	Perl subroutines: setLineBreak; treeRead; duplStrip; pruneTree; cladeClean; cleanTree
#
# Usage: treePruner.pl -t<filename> [-p<b|c|d>] [-l<filename>] [-s] [-v]
#   options: -l<filename> = text file containing names of taxa to be pruned one per line 
#							(can automatically read output from STRindexer.pl)
#			 -p<b|c|d> = pruned input trees to their common taxon set (c -- default), of a specific set of taxa (d), or both (b)
#			 -s = maintain any support values in trees
#            -t<filename> = file containing trees in NEXUS format
#            -v = verbose output
#            -h = print this message and quit

use strict;
use File::Basename;

# Global variables
	# I/O parameters
		my ($treeFile, $outFile, $taxonFile);
			my ($dataSource, $dir, $ext);
		my $block = 1;
			my $treeBlock = 0;

	# Pruning variables
		my @treeSpecies;
		my $keepSupport = 0;
		my $pruneType = "common";	# Options are "common" (prune to taxon on all trees), "directed" (prune specific taxa), and "both"
			my %pruneTaxon;
		my $emptySet = 0;
		
	# Tree variables
		my (%treeName, %treeDescription, %rootingStatus, %prunedTree);
		my $unrootedFlag = 0;
		my $defaultRooting = 0;	# Default is for trees to be unrooted
		my $SupportFlag = 0;
		my @allTaxa;
	
	# Miscellaneous
		my $verbose = 0;
		my $debug = 0;

# Process flags
	for (my $i = 0; $i <= $#ARGV; $i++)
		{
		if ($ARGV[$i] eq "-pb")
			{
			$pruneType = "both";
			}
		elsif ($ARGV[$i] eq "-pc")
			{
			$pruneType = "common";
			}
		elsif ($ARGV[$i] eq "-pd")
			{
			$pruneType = "directed";
			}
		elsif ($ARGV[$i] =~ /-l([\w+\.?\w+?]*)/)
			{
			$taxonFile = $1;
			}
		elsif ($ARGV[$i] eq "-s")
			{
			$keepSupport = 1;
			}
		elsif ($ARGV[$i] =~ /-t([\w+\.?\w+?]*)/)
			{
			$treeFile = $1;
			($dataSource, $dir, $ext) = fileparse($treeFile, qr/\.\D.*/);
			$outFile = $dataSource . "_pruned.tre";
			}
		elsif ($ARGV[$i] eq "-v")
			{
			$verbose = 1;
			}
		elsif ($ARGV[$i] eq "-x")
			{
			$debug = $verbose = 1;
			}
		elsif ($ARGV[$i] eq "-h")
			{
			print "Usage: treePruner.pl -t<filename> [-p<b|c|d>] [-l<filename>] [-s] [-v]\n";
			print "Options: -l<filename> = text file containing names of taxa to be pruned one per line\n";
			print "                        (can automatically read output from STRindexer.pl)\n";
			print "         -p<b|c|d> = pruned input trees to their common taxon set (c -- default), of a specific set of taxa (d), or both (b)\n";
			print "         -s = maintain any support values in trees\n";
			print "         -t<filename> = file containing trees in NEXUS format\n";
			print "                        (default = treefile.tre)\n";
			print "         -v = verbose output\n";
			print "         -h = print this message and quit\n";
			exit(0);
			}
		else
			{
			print "Don't understand argument: $ARGV[$i]\n";
			print "Usage: treePruner.pl -t<filename> [-p<b|c|d>] [-l<filename>] [-s] [-v]\n";
			exit(1); 
			}
		}

# Check for necessary information
	die "ERROR: Must give name of nexus-formatted treefile.\n" unless (defined $treeFile);
	die "ERROR: Must give name of text file containing names of taxa to be pruned.\n" if (not defined $taxonFile and ($pruneType eq "directed" or $pruneType eq "both"));

# Get tree descriptions
	print "\nReading in tree descriptions from file $treeFile ...\n";
		my @allTrees = treeRead($treeFile);
		die "No trees present in $treeFile" if (not @allTrees);

# Get taxa to be pruned as needed
	if ($pruneType eq "directed" or $pruneType eq "both")
		{
		print "\nReading in taxa to be pruned from file $taxonFile ...\n";
		setLineBreak($taxonFile);
		open (TAXA, "<$taxonFile") or die "Cannot open file containing names of taxa to be pruned $taxonFile\n";
			while (<TAXA>)
				{
				chomp;
				next unless ($_);
					# Safeties for STRindexer.pl output
						next if ($_ =~ /\(C\*\)$/);
						last if ($_ =~ /^=============/);
				
				my @taxonLine = split(/\s+/, $_);
					$pruneTaxon{shift(@taxonLine)} = 1;
				}
		close TAXA;
		}

# Perform pruning
	print "\nPruning trees to common taxon set ...\n" if ($pruneType eq "common");
	print "\nPruning trees of user-input taxa ...\n" if ($pruneType eq "directed");
	print "\nPruning trees to common taxon set and of user-input taxa ...\n" if ($pruneType eq "both");

	my $pruneZero = time;
		pruneTree(@allTrees);
		printf "\tTime taken: %s seconds\n", time - $pruneZero;
	
# Print results!
	print "\nPrinting results to file $outFile ...\n";
	open (TREE, ">$outFile") or die "Cannot write to treefile, $outFile\n";
		print TREE "#nexus\n\nbegin trees;\n";
		foreach my $treeID (@allTrees)
			{
			print TREE "\t";
			if ($prunedTree{$treeID} eq "()")	# Safety if tree pruned to empty set
				{
				print TREE "[";
				$emptySet++;
				}
			print TREE "tree $treeName{$treeID} = ";
				if ($rootingStatus{$treeID} == 1)
					{
					print TREE "[&R]";
					}
				else
					{
					print TREE "[&U]";
					}
			print TREE " $prunedTree{$treeID};";
			print TREE "]" if ($prunedTree{$treeID} eq "()");
			print TREE "\n";
			}
		print TREE "end;\n";
	close TREE;
	printf "\tWARNING: $emptySet of %s input trees pruned to empty set\n", scalar(@allTrees) if ($emptySet);
	
# Subroutines used in the script

sub treeRead	# Subroutine to read in tree descriptions and derive taxon names; accounts for any translation tables
	{
	my $treeFile = shift;
	my $treeNum = 0;

	setLineBreak($treeFile);
	open (TREES, "<$treeFile") or die "Cannot open nexus-formatted treefile $treeFile\n";
		my (@allTrees, $rawTreeDescrip, $treeDescrip, $treeFlag, $treeCount, $rooting);
		my (@taxonList, %taxonPresent);
		my (%TTable, $translation, $translateFlag, $maxTaxonID, $transText);
		
		while (<TREES>)
			{
			chomp;
			$_ =~ s/^\s+//;		# Strip (any) leading white space
			if ($_ =~ m/begin trees;/i)		# Clear variables for each tree block
				{
				undef %TTable;
				$translation = $translateFlag = $treeFlag = $maxTaxonID = $treeCount = 0;
				$treeBlock++;
				}
			if ($_ =~ m/translate/i)	#Read translation table of taxon names (if any)
				{
				$translateFlag = $translation = 1;
				$transText = "";
				print "Reading translation table ...\n" if ($verbose);
				}
			if ($translateFlag)
				{
				unless ($_ =~ m/translate/i)
					{
					$transText .= $_;
					if ($_ =~ m/;/)
						{
						$translateFlag = 0;
						$transText =~ s/;$//;	#Strip trailing semi-colon
						my @transEntries = split(/,/, $transText);
						foreach my $entry (@transEntries)
							{
							$entry =~ s/^\s+//;		#Strip (any) leading white space
							my ($taxonID, $taxonName) = split(/\s/,$entry);
							$maxTaxonID = $taxonID if ($taxonID > $maxTaxonID);
							$taxonName =~ s/\s+//g;
							$TTable{$taxonID} = $taxonName;
							}
						}
					}
				}
			if ($_ =~ /tree/i and $_ =~ /=/)
				{
				unless ($_ =~ /^>/)		#Ignore (any) commented lines
					{
					$rawTreeDescrip = $treeDescrip = "";
					unless ($block == 0 and $treeBlock > 1)
						{
						$treeFlag = 1;
						}
					}
				}
			if ($treeFlag)
				{
				$rawTreeDescrip .= $_;
				if ($_ =~ m/:/)
					{
					$SupportFlag = 1;
					}
				if ($_ =~ m/;/)
					{
					$treeCount++;
					print "\nTree block: $treeBlock\tTree number: $treeCount\n" if ($verbose);
					print "\tTranslating taxon names ...\n" if ($verbose and $translation);
					# Determine if trees are rooted or not (default for no information is not rooted)
						if ($rawTreeDescrip =~ /\[&R\]/)
							{
							$rooting = 1;
							}
						elsif ($rawTreeDescrip =~ /\[&U\]/)
							{
							$rooting = 0;
							$unrootedFlag = 1;
							}
						else
							{
							$rooting = $defaultRooting;
							$unrootedFlag = 1 if ($defaultRooting == 0);
							}
					# Remove any embedded comments
						if ($rawTreeDescrip =~ /\[/)
							{
							for (my $position = 0; $position < length($rawTreeDescrip); $position++)
								{
								if (substr($rawTreeDescrip, $position, 1) eq "[")
									{
									$position++ until (substr($rawTreeDescrip, $position, 1) eq "]");
									$position++;
									}
								else
									{
									$treeDescrip .= substr($rawTreeDescrip, $position, 1);
									}
								}
							}
						else
							{
							$treeDescrip = $rawTreeDescrip;
							}
					$treeFlag = 0;
					my $position = 0;
					until (substr($treeDescrip, $position, 1) eq "(")		# Avoid anything before tree description proper
						{
						$position++;
						}
					my $tree = substr($treeDescrip, $position);
						$tree =~ s/;\s*$//;
						$tree =~ s/\s+//g;
					my ($name) = (split /=/, $treeDescrip, 0);		# Get name of tree, if any
						$name =~ s/\s*\w*tree\s+//i;
						$name =~ s/\s*\*\s*//i;
						$name =~ s/\s+$//g;
#						$treeNames{$treeCount} = $name;
					my $openCount = ($tree =~ tr/(//);
					my $closedCount = ($tree =~ tr/)//);
					if ($openCount == $closedCount)		# Tree is balanced; process
						{
						$position = length($tree) - 1;
						while ($position >= 0)
							{
							my $IDlength = 1;
							if (substr($tree, $position, 1) eq "(" or substr($tree, $position, 1) eq ")" or substr($tree, $position, 1) eq "," or substr($tree, $position, 1) eq ":")		#Tree structure; skip to next character
								{
								$position--;
								}
							else	# Read in taxonID
								{
								until (substr($tree, $position, 1) eq "(" or substr($tree, $position, 1) eq ")" or substr($tree, $position, 1) eq "," or substr($tree, $position, 1) eq ":")
									{
									$position--;
									$IDlength++;
									}
								unless (substr($tree, $position, 1) eq ":")		# Avoid reading branch lengths as taxon names
									{
									my $readName = substr($tree, $position+1, $IDlength-1);
									if ($translation)	# Translate taxon names
										{
										if ($TTable{$readName})
											{
											substr($tree, $position+1, $IDlength-1) = $TTable{$readName};		#Translate name
											unless ($taxonPresent{$TTable{$readName}})
												{
												push @taxonList, $TTable{$readName};
												$taxonPresent{$TTable{$readName}} = 1;
												}
											}
										else
											{
											print "\t\tUnidentified taxon ".$readName." in tree #".$treeCount." of tree block #".$treeBlock.". Replaced with xxx.\n";
											substr($tree, $position+1, $IDlength-1) = "xxx";		#Translate name
											unless ($taxonPresent{"xxx"})
												{
												push @taxonList, "xxx";
												$taxonPresent{"xxx"} = 1;
												}
											}
										}
									else	#Enter name as read from tree, unless already present
										{
										unless ($taxonPresent{$readName})
											{
											push @taxonList, $readName;
											$taxonPresent{$readName} = 1;
											}
										}
									}
								}
							}
						$treeNum++;
							my $treeID = "treeID_$treeNum";
						
						push @allTrees, $treeID;
							$treeDescription{$treeID} = $tree;
							$rootingStatus{$treeID} = $rooting;
							$treeName{$treeID} = $name;
						print "\tTree description:\n\t\t$tree\n" if ($verbose);
						}
					else	#Tree is unbalanced; reject
						{
						print "\tTree #".$treeCount." of tree block #".$treeBlock." is unbalanced. Not processed.\n";
						}
					}
				}
			}
	close TREES;
	
	# Process taxon list
		@allTaxa = sort(duplStrip(@taxonList));
	
	return @allTrees;
	}

sub setLineBreak	# Check line breaks of input files and set input record separator accordingly
	{
	my $inFile = shift;
	$/ ="\n";
	open (IN, "<$inFile") or die "Cannot open $inFile to check form of line breaks.\n";
		while (<IN>)
			{
			if ($_ =~ /\r\n/)
				{
				print "DOS line breaks detected ...\n" if ($verbose);
				$/ ="\r\n";
				last;
				}
			elsif ($_ =~ /\r/)
				{
				print "Mac line breaks detected ...\n" if ($verbose);
				$/ ="\r";
				last;
				}
			else
				{
				print "Unix line breaks detected ...\n" if ($verbose);
				$/ ="\n";
				last;
				}
			}
	close IN;
	}

sub duplStrip	# Subroutine to strip duplicate entries from an array
	{
	my @duplArray = @_;
	my %dupl;		# Remove duplicate taxa
		my @stripArray = grep(!$dupl{$_}++, @duplArray);
		@stripArray = sort {$a cmp $b} @stripArray;
	return @stripArray;
	}

sub pruneTree
	{
	my @treeList = @_;
	my (%taxonCount, @taxonList, @finalList);

	# Get taxon names on all trees
		print "Checking taxon names ...\n" if ($debug);	
		foreach my $treeID (@treeList)
			{
			my $tree = $treeDescription{$treeID};
				$tree =~ s/:\d+\.?\d*e?\-?\d*//g unless ($keepSupport);
				$tree =~ s/:\d+\.?\d*e?\-?\d*//g;
				$tree =~ s/\(//g;
				$tree =~ s/\)//g;
				print "\t$tree\n" if ($debug);

			foreach my $taxon (split (/,/, $tree))
				{
				$taxon =~ s/\s+/_/g;
				$taxonCount{$taxon}++;
				push @taxonList, $taxon if ($taxonCount{$taxon} == 1);
				}
			}
	
	# Process trees, keeping only those taxa on all trees
		print "Pruning taxa ...\n" if ($debug);	
		foreach my $treeID (@treeList)
			{
			my $tree = $treeDescription{$treeID};
				$tree =~ s/:\d+\.?\d*e?\-?\d*//g unless ($keepSupport);
				print "\t$tree\n" if ($debug);
			undef @treeSpecies;
			foreach my $taxon (@taxonList)
				{
				if ($pruneType eq "common" and $taxonCount{$taxon} == scalar(@treeList))
					{
					push @treeSpecies, $taxon;
					}
				elsif ($pruneType eq "directed" and not defined $pruneTaxon{$taxon})
					{
					push @treeSpecies, $taxon;
					}
				elsif ($pruneType eq "both" and $taxonCount{$taxon} == scalar(@treeList) and not defined $pruneTaxon{$taxon})
					{
					push @treeSpecies, $taxon;
					}
				else	# Remove taxon along with any associated branch length
					{
					$tree =~ s/\($taxon(\:\d+\.?\d*e?\-?\d*)?\)/\(\)/g;	# Prefix = ( suffix = )
					$tree =~ s/\($taxon(\:\d+\.?\d*e?\-?\d*)?\,/\(\,/g;	# Prefix = ( suffix = ,
					$tree =~ s/\,$taxon(\:\d+\.?\d*e?\-?\d*)?\)/\,\)/g;	# Prefix = , suffix = )
					$tree =~ s/\,$taxon(\:\d+\.?\d*e?\-?\d*)?\,/\,\,/g;	# Prefix = , suffix = ,
					}
				}
			print "\t$tree\n" if ($debug);
			
			$prunedTree{$treeID} = cleanTree($tree);
#			my $finalTree = cleanTree($tree);
#			push @finalList, $finalTree;
			}

#	return @finalList;
	}

sub cleanTree	# Subroutine to remove extraneous markers from tree description due to pruning taxa
	{
	my $dirtyTree = shift;
	print "\t\tCleaning tree ...\n" if ($debug);

	my $changeCount = 1;
	until ($changeCount == 0)	# Remove most empty clade markers from pruned ST description
		{
		$changeCount = $dirtyTree =~ s/,,/,/g;
		$changeCount += $dirtyTree =~ s/\(\)(:\d+\.?\d*e?\-?\d*)?//g;
		$changeCount += $dirtyTree =~ s/\(,\)(:\d+\.?\d*e?\-?\d*)?//g;
		$changeCount += $dirtyTree =~ s/\(,\(/\(\(/g;
		$changeCount += $dirtyTree =~ s/\),\)/\)\)/g;
		$changeCount += $dirtyTree =~ s/,\)/\)/g;
		$changeCount += $dirtyTree =~ s/\(,/\(/g;

		foreach my $species (@treeSpecies)	# Check for "clades" consisting of a single species: "(species:support)"
			{
			if ($keepSupport and $dirtyTree =~ /(\($species\:\d+\.?\d*e?\-?\d*\)\:\d+\.?\d*e?\-?\d*)/)
				{
				my $replaceText = $1;
					$replaceText =~ s/^\(//;
					$replaceText =~ s/\)\:\d+\.?\d*e?\-?\d*$//;
					
				$dirtyTree =~ s/\($species\:\d+\.?\d*e?\-?\d*\)\:\d+\.?\d*e?\-?\d*/$replaceText/;
				$changeCount++;
				}
			if ($dirtyTree =~ /\($species\)/)
				{
				$dirtyTree =~ s/\($species\)/$species/;
				$changeCount++;
				}
			}
		}
	print "\t\t\tDirty tree: $dirtyTree\n" if ($debug);

	# Fix for trees synonymized to a single taxon
		if (substr($dirtyTree, 0, 1) ne "(" and substr($dirtyTree, length($dirtyTree), 1) ne ")")
			{
			$dirtyTree = "(".$dirtyTree.")";
			}
			
	my $cleanTree = cladeClean($dirtyTree);
		$cleanTree =~ s/\:\d+\.?\d*e?\-?\d*$//;	# Strip trailing support value
	print "\t\t\tClean tree: $cleanTree\n" if ($debug);
	return $cleanTree;
	}

sub cladeClean	# Subroutine to remove "empty" clades from tree description
	{
	my $workTree = shift;
		print "\t\t\t\tWorking tree: $workTree\n" if ($debug);

	my ($position, $cladeID, $currentClade, $offset) = (0, 0, 0, 0);
	my $oldCurrent;
	my (%parentClade, %cladeDescrip, %cladeSupport, %start, %stop);
		$start{0} = 0;
		$stop{0}= length($workTree);
	my (%present, %deletedClade);
	my @delPositions;	

	# Get clade descriptions
		while ($position < length($workTree))
			{
			$offset = 1;
			if (substr($workTree, $position, 1) eq "(")		# New clade; go one level deeper
				{
				$cladeID++;
				$parentClade{$cladeID} = $currentClade;
				$currentClade = $cladeID;
				$start{$cladeID} = $position;
				$deletedClade{$cladeID} = 0;
				}
			elsif (substr($workTree, $position, 1) eq ")")	# Clade finished; clean up and go up one level
				{
				$oldCurrent = $currentClade;
				$currentClade = $parentClade{$oldCurrent};
				$stop{$oldCurrent} = $position;

				# Get clade count to check for duplicate clades
					$cladeDescrip{$oldCurrent} = substr($workTree, $start{$oldCurrent}, $stop{$oldCurrent} - $start{$oldCurrent} + 1);
						$cladeDescrip{$oldCurrent} =~ s/:\d+\.?\d*e?\-?\d*//g;
						$cladeDescrip{$oldCurrent} =~ s/\(//g;
						$cladeDescrip{$oldCurrent} =~ s/\)//g;
						
					$present{$cladeDescrip{$oldCurrent}}++;
					print "\t\t\t\t\tClade $oldCurrent: $cladeDescrip{$oldCurrent} ($present{$cladeDescrip{$oldCurrent}})\n" if ($debug);
					$deletedClade{$oldCurrent} = 1 if ($present{$cladeDescrip{$oldCurrent}} > 1);	# Clade is identical to one nested within it: delete
				}
			elsif (substr($workTree, $position, 1) eq ":")	# Branch lengths present; record as nodal support
				{
				until (substr($workTree, $position+$offset, 1) eq "(" or substr($workTree, $position+$offset, 1) eq ")" or substr($workTree, $position+$offset, 1) eq "," or $position+$offset >= length($workTree))
					{
					$offset++;
					}
				if (substr($workTree, $position-1, 1) eq ")")	# Support value pertains to a clade
					{
					$cladeSupport{$oldCurrent} = substr($workTree, $position+1, $offset-1);
					}
				$position += $offset-1;
				}
			elsif (substr($workTree, $position, 1) eq ",")
				{
				}
			else	#Determine membership of clade
				{
				until (substr($workTree, $position+$offset, 1) eq "(" or substr($workTree, $position+$offset, 1) eq ")" or substr($workTree, $position+$offset, 1) eq ",")
					{
					$offset++;
					}
				$position += $offset-1;
				}
			$position++;
			}
		
	# Reparse tree and print parentheses and clade supports unless clade is marked for deletion
		my $growTree;
		$position = $cladeID = 0;
		while ($position < length($workTree))
			{
			$offset = 1;
			if (substr($workTree, $position, 1) eq "(")		# New clade; go one level deeper
				{
				$cladeID++;
				$growTree .= "(" unless ($deletedClade{$cladeID});
				$currentClade = $cladeID;
				}
			elsif (substr($workTree, $position, 1) eq ")")	# Clade finished; clean up and go up one level
				{
				$oldCurrent = $currentClade;
				unless ($deletedClade{$currentClade})
					{
					$growTree .= ")";
					$growTree .= ":$cladeSupport{$currentClade}" if ($keepSupport and defined $cladeSupport{$currentClade});
					}
				$currentClade = $parentClade{$oldCurrent};
				}
			elsif (substr($workTree, $position, 1) eq ":")	# Branch lengths present; record as nodal support
				{
				until (substr($workTree, $position+$offset, 1) eq "(" or substr($workTree, $position+$offset, 1) eq ")" or substr($workTree, $position+$offset, 1) eq "," or $position+$offset >= length($workTree))
					{
					$offset++;
					}
				unless (substr($workTree, $position-1, 1) eq ")")	# Support value pertains to a terminal taxon
					{
					$growTree .= ":" . substr($workTree, $position+1, $offset-1) if ($keepSupport);
					}
				$position += $offset-1;
				}
			elsif (substr($workTree, $position, 1) eq ",")
				{
				$growTree .= ",";
				}
			else	#Determine membership of clade
				{
				$growTree .= substr($workTree, $position, 1);
				until (substr($workTree, $position+$offset, 1) eq "(" or substr($workTree, $position+$offset, 1) eq ")" or substr($workTree, $position+$offset, 1) eq "," or substr($workTree, $position+$offset, 1) eq ":")
					{
					$growTree .= substr($workTree, $position+$offset, 1);
					$offset++;
					}
				$position += $offset-1;
				}
			$position++;
			}

	$growTree =~ s/\s//g;
	print "\t\tEmpty clean: $growTree\n" if ($debug);
	return $growTree;
	}

# Version history
#
#	v1.0 (June 16, 2005)
#		- initial release
